package tesuji.games.go;

import java.util.Stack;

import tesuji.games.general.Cursor;
import tesuji.games.go.BoardMarker;
import tesuji.games.go.GoArray;
import tesuji.games.general.FlyWeight;

/**
 	This class acts like a (usually small) list of integers.
 	The only way to add items is through the 'push' method.
 	But otherwise it acts just like an array of integers.

 	<br><br>Creation date: (17-May-01 1:52:59 PM)<br><br>

	Product: Tesuji Software Go Library.<br><br>

	<font color="#CC6600"><font size=-1>
	Copyright (c) 2001 Tesuji Software B.V.<br>
	All rights reserved.<br><br>

	Permission is hereby granted, free of charge, to any person obtaining a
	copy of this software and associated documentation files (the
	"Software"), to deal in the Software without restriction, including
	without limitation the rights to use, copy, modify, merge, publish,
	distribute, and/or sell copies of the Software, and to permit persons
	to whom the Software is furnished to do so, provided that the above
	copyright notice(s) and this permission notice appear in all copies of
	the Software and that both the above copyright notice(s) and this
	permission notice appear in supporting documentation.<br><br>

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
	COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
	IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
	CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.<br><br>

	Except as contained in this notice, the name of a Tesuji Software
	shall not be used in advertising or otherwise to promote the sale, use
	or other dealings in this Software without prior written authorization
	of Tesuji Software.<br><br>
	<font color="#00000"><font size=+1>

	@author: Mark Boon
*/
public class IntStack
	implements FlyWeight
{
	private static BoardMarker marker = new BoardMarker();
	private static IntStack toDoList = new IntStack();
		
	private int index = 0;
	private int[] stack;
	
	private static Stack iteratorStack = new Stack();

	 // This is the stack that owns the object when it gets
	 // created and to which it gets recycled after use.
	private Stack owner;

	/**
	 	Default constructor. A GoArray is used so that we're
	 	sure the capacity is sufficient to store enough items
	 	to fill a board.
	
	 	<br><br>Creation date: (17-May-01 2:01:22 PM)<br><br>
	*/
	public IntStack()
	{
		stack = GoArray.createIntegers();
	}
	
	/**
	 	IntStack constructor with a given capacity.
	
	 	<br><br>Creation date: (17-May-01 2:01:22 PM)<br><br>
	*/
	public IntStack(int size)
	{
		stack = new int[size];
	}
	
	public void push(int value)
	{
		stack[index++] = value;
	}
	
	public int pop()
	{
		return stack[--index];
	}
	
	public int peek()
	{
		return stack[index-1];
	}
	
	public int peek(int offset)
	{
		return stack[index-offset-1];
	}
	
	/**
	 	Remove an item from the list. This may cause the
	 	order of the items in the list to change, so don't
	 	use this method if the order of the items is important.
	
	 	<br><br>Creation date: (17-May-01 2:20:11 PM)<br><br>
	
	 	@param: int the value to remove from the list.
	*/
	public void remove(int value)
	{
		index--;
		int last = stack[index];
		if (value!=last)
		{
			// If it's not the last item in the list, loop until
			// it's found and store the last item in the place
			// of the item to remove.
			for (int i=index-1; i>=0; i--)
			{
				if (stack[i]==value)
				{
					stack[i] = last;
					return;
				}
			}
		}
		else
			return;

		// The item was not found in the list!
		index++;
	}

	public boolean isEmpty()
	{
		return (index==0);
	}
	
	public int getSize()
	{
		return index;
	}

	/**
	 	This method clears the list and then starts to fill
	 	it with the coordinates of all the four-way-connected
	 	members of the same value in the board array.<br><br>

	 	For example this can be used to get the coordinates
	 	of all the stones that belong to the same chain.
	
	 	<br><br>Creation date: (17-May-01 2:12:02 PM)<br><br>
	
	 	@param: int startXY is the starting point.
	 	@param: byte[] board is the array with values.
	*/
	public final void getMembers(int startXY, byte[] board)
	{
		clear();
		int xy;
		byte color = board[startXY];
		toDoList.clear();
		toDoList.push(startXY);
		do
		{
			xy = toDoList.pop();
			if (marker.notSet(xy))
			{
				push(xy);
				marker.set(xy);
				int left = GoArray.left(xy);
				int right = GoArray.right(xy);
				int above = GoArray.above(xy);
				int below = GoArray.below(xy);
				if (board[left]==color && marker.notSet(left))
					toDoList.push(left);
				if (board[right]==color && marker.notSet(right))
					toDoList.push(right);
				if (board[above]==color && marker.notSet(above))
					toDoList.push(above);
				if (board[below]==color && marker.notSet(below))
					toDoList.push(below);
			}	
		}
		while (!toDoList.isEmpty());
	}

	/**
	 	Get a Cursor object that can be used to iterate
	 	over all the values in the IntStack.

	 	Make sure to call 'recycle' when you're done with
	 	it for efficiency.
	
	 	<br><br>Creation date: (17-May-01 2:01:22 PM)<br><br>
	*/
	public Cursor getIterator()
	{
		if (iteratorStack.size()==0)
			return new Iterator(this);

		Iterator iterator = (Iterator) iteratorStack.pop();
		iterator.setIntStack(this); // Hmmm... this 'starts' the iterator.
		return iterator;
	}
	
	// This is an internal class that implements an iterator
	// over all the items currently on the stack. Using this
	// iterator is rather less efficient than a plain loop:
	//
	//	for (int i=0; i<intStack.getSize(); i++)
	//		nextValue = intStack.get(i);
	//
	// But occasionally it does have its uses.
	// Make sure to recycle it after use for more efficiency.
	
	private class Iterator
		implements Cursor
	{
		int index ;
		IntStack intStack;
		
		public Iterator(IntStack stack)
		{
			intStack = stack;
		}

		public void setIntStack(IntStack stack)
		{
			intStack = stack;
			index = intStack.getSize();
		}
		
		public boolean continues()
		{
			return (index>0);
		}
		
		public int next()
		{
			return intStack.stack[--index];
		}
		
		public void start()
		{
			index = intStack.getSize();
		}

		public void recycle()
		{
			iteratorStack.push(this);
		}
	}

	/**
	 	IntStack constructor. The ownerStack is used so that
	 	the method 'recycle' can be used for deallocation.
	
	 	<br><br>Creation date: (17-May-01 2:01:22 PM)<br><br>
	*/
	public IntStack(Stack ownerStack)
	{
		owner = ownerStack;
		stack = GoArray.createIntegers();
	}

	public void add(IntStack otherStack)
	{
		for (int i=otherStack.getSize()-1; i>=0; i--)
			push(otherStack.get(i));
	}

	/**
	 	Clear the contents of the IntStack object.
	 	It will now have size 0 and contain 0 items.
	
	 	<br><br>Creation date: (17-May-01 2:09:02 PM)<br><br>
	*/
	public void clear()
	{
		index = 0;
		marker.getNewMarker(); // Because of getMoreMembers.
	}

	public final int get(int index)
	{
		return stack[index];
	}

	public final int[] getList()
	{
		return stack;
	}

	/**
		This method does the same as getMembers but it doesn't
		clear the list first, so the members get added to the
		existing members. It also doesn't clear the board-markers
		to prevent the same coordinates to be entered more than
		once. This implies that this method can only be used
		after getMembers has been used first.
	
	 	<br><br>Creation date: (17-May-01 2:12:02 PM)<br><br>
	
	 	@param: int startXY is the starting point.
	 	@param: byte[] board is the array with values.
	*/
	public final void getMoreMembers(int startXY, byte[] board)
	{
		int xy;
		byte color = board[startXY];
		toDoList.clear();
		toDoList.push(0);		
		toDoList.push(startXY);
		while ((xy=toDoList.pop())!=0)
		{
			if (marker.notSet(xy))
			{
				push(xy);
				marker.set(xy);
				int left = GoArray.left(xy);
				int right = GoArray.right(xy);
				int above = GoArray.above(xy);
				int below = GoArray.below(xy);
				if (board[left]==color && marker.notSet(left))
					toDoList.push(left);
				if (board[right]==color && marker.notSet(right))
					toDoList.push(right);
				if (board[above]==color && marker.notSet(above))
					toDoList.push(above);
				if (board[below]==color && marker.notSet(below))
					toDoList.push(below);
			}	
		}
	}

	public final void recycle()
	{
		if (owner!=null)
			owner.push(this);
	}

	/**
		Setting the size of the number of items on the stack.
		Should only be used if the current size is greater than
		the new size and can be used to 'pop' more than one item.
		Use with care.
	
	 	<br><br>Creation date: (17-May-01 2:01:22 PM)<br><br>
	*/
	public void setSize(int newSize)
	{
		index = newSize;
	}
}